import os
from typing import Iterable
from typing import Mapping
from typing import TypeVar
from typing import Union

T = TypeVar("T")
Scalar = Union[int, float, bool, str]
IterableOrSingle = Union[Iterable[T], T]
PathLike = Union[str, bytes, os.PathLike]

Section = Mapping[str, IterableOrSingle[Scalar]]
InifixParsable = Mapping[str, Union[Section, Scalar]]



import warnings
from functools import wraps
from types import FunctionType
from typing import Dict


def future_positional_only(positions2names: Dict[int, str]):
    """Warn users when using a future positional-only argument as keyword.
    Note that positional-only arguments are available from Python 3.8
    See https://www.python.org/dev/peps/pep-0570/
    """

    def outer(func: FunctionType):
        @wraps(func)
        def inner(*args, **kwargs):
            for no, name in sorted(positions2names.items()):
                if name not in kwargs:
                    continue
                msg = (
                    f"using the {name!r} argument as keyword (on position {no}) "
                    "is deprecated and will stop working in a future release. "
                    "Pass the argument as positional to supress this warning."
                )
                warnings.warn(msg, FutureWarning, stacklevel=2)
            return func(*args, **kwargs)

        return inner

    return outer



import argparse
import os
import re
import sys
import warnings
from difflib import unified_diff
from typing import Optional
from typing import Sequence

from inifix.io import load

NAME_COL_MINSIZE = 10
COMMENT_COL = 80
PADDING_SIZE = 4


def _ljust(s: str, width: int) -> str:
    """
    A more flexible str.ljust
    Guarantees the return value has at least one trailing whitespace char,
    even at the cost of going over the specified width.
    """
    justified_name = s.ljust(width)
    if len(s) < width:
        return justified_name
    else:
        return justified_name + " "


def _normalize_whitespace(s: str) -> str:
    return re.sub(r"\s", " ", s).strip()


def iniformat(data: str, *, name_column_size: Optional[int] = None) -> str:
    lines = data.splitlines()
    contents = []
    comments = []
    parameters = []
    values = []
    for line in lines:
        content, _, comment = line.partition("#")
        content = _normalize_whitespace(content)
        contents.append(content)
        comments.append(_normalize_whitespace(comment))
        if not content.startswith("[") and content != "":
            parameter, *value = content.split()
            parameters.append(parameter)
            values.append(value)

    if not parameters:
        return "\n".join(lines)

    max_name_size = max(len(parameter) for parameter in parameters)

    if name_column_size is None:
        padded_name_col_size = (
            max(max_name_size + PADDING_SIZE, NAME_COL_MINSIZE) + PADDING_SIZE
        )
    else:
        padded_name_col_size = name_column_size

        if max_name_size >= name_column_size:
            long_parameters = [p for p in parameters if len(p) >= name_column_size]
            warnings.warn(
                "The following parameters\n"
                + "\n".join(long_parameters)
                + f"\nare longer than user specified column length ({name_column_size}). "
                + "Column length will be adjusted regardless.",
            )

    new_lines = []
    parameter_idx = 0
    for content, comment in zip(contents, comments):
        new_line = ""
        offset = 0
        if content:
            if content.startswith("["):
                new_line += f"{content}\n"
            else:
                new_line += _ljust(parameters[parameter_idx], padded_name_col_size)
                new_line += "  ".join(values[parameter_idx])

                offset = COMMENT_COL - len(new_line)
                parameter_idx += 1
        comm = f"# {comment}"
        new_line += comm.rjust(offset + len(comm)) if comment else ""
        new_lines.append(new_line)
    res = "\n".join(new_lines)
    res = re.sub("\n+", "\n", res)
    res = re.sub(r"(.?)\n\[", r"\1\n\n\n[", res)
    res = re.sub(r"\]\n+", r"]\n\n", res)
    # ensure there's exactly one newline at the EOF
    res = res.rstrip("\n")
    res += "\n"
    return res


def main(argv: Optional[Sequence[str]] = None) -> int:
    parser = argparse.ArgumentParser()
    parser.add_argument("files", nargs="+")
    parser.add_argument(
        "--diff",
        action="store_true",
        help="Print the unified diff to stdout instead of editing files inplace",
    )
    parser.add_argument(
        "--name-column-size", type=int, help="Fixed length of the parameter column"
    )

    args = parser.parse_args(argv)

    retv = 0

    for file in args.files:
        if not os.path.isfile(file):
            print(f"Error: could not find {file}", file=sys.stderr)
            retv += 1
            continue
        try:
            load(file)
        except ValueError as exc:
            print(f"Error: {exc}", file=sys.stderr)
            retv += 1
            continue

        with open(file) as fh:
            data = fh.read()

        fmted_data = iniformat(data, name_column_size=args.name_column_size)

        if fmted_data == data:
            print(f"{file} is already formatted", file=sys.stderr)
            continue
        else:
            retv += 1

        if args.diff:
            for line in unified_diff(
                data.splitlines(), fmted_data.splitlines(), fromfile=file
            ):
                if sys.version_info >= (3, 9):
                    line = line.removesuffix("\n")
                elif line.endswith("\n"):
                    line = line[:-1]
                print(line)
            print("\n")
        else:
            print(f"Fixing {file}", file=sys.stderr)
            try:
                with open(file, "w") as fh:
                    fh.write(fmted_data)
            except OSError:
                print(f"Error: could not write to {file}", file=sys.stderr)
                retv += 1
                continue

    return retv


if __name__ == "__main__":  # pragma: no cover
    sys.exit(main())



from typing import TextIO
from typing import Union

from inifix._deprecation import future_positional_only
from inifix._typing import PathLike
from inifix.iniconf import InifixConf


@future_positional_only({0: "dict_or_path_or_buffer"})
def load(dict_or_path_or_buffer: Union[dict, PathLike, TextIO]) -> InifixConf:
    return InifixConf(dict_or_path_or_buffer)


@future_positional_only({0: "conf", 1: "file_descriptor"})
def dump(conf: dict, file_descriptor: TextIO) -> None:
    """Write a `conf` dict to a file.

    `file_descriptor` can represent a file path or an object with a `write` method.
    """
    conf = InifixConf(conf)
    conf.write(file_descriptor)



import re
from collections import defaultdict
from typing import Callable
from typing import List
from typing import TextIO
from typing import Tuple
from typing import Union

from more_itertools import always_iterable
from more_itertools import mark_ends

from inifix._typing import InifixParsable
from inifix._typing import IterableOrSingle
from inifix._typing import PathLike
from inifix._typing import Scalar
from inifix._typing import Section
from inifix.enotation import ENotationIO
from inifix.validation import validate_inifile_schema

SECTION_REGEXP = re.compile(r"\[.+\]\s*")


def bool_caster(s: str) -> bool:
    s = s.lower()
    if s in ("true", "t"):
        return True
    elif s in ("false", "f"):
        return False
    raise ValueError


def str_caster(s: str) -> str:
    if re.match(r"^'.*'$", s) or re.match(r'^".*"$', s):
        return s[1:-1]
    else:
        return s


class InifixConf(dict):
    def __init__(
        self, dict_or_path_or_buffer: Union[InifixParsable, PathLike, TextIO]
    ) -> None:
        """
        Parse a .ini configuration from a file, or a dict.

        Parameters
        ----------
        dict_filepath_or_buffer: dict, os.PathLike or str or a readable file handle.
        """
        if isinstance(dict_or_path_or_buffer, dict):
            validate_inifile_schema(dict_or_path_or_buffer)
            super().__init__(dict_or_path_or_buffer)
            return
        self._from_file(dict_or_path_or_buffer)
        validate_inifile_schema(self)

    def _from_file(self, filepath_or_buffer: Union[PathLike, TextIO]) -> None:

        _dict: InifixParsable = defaultdict(dict)

        try:
            data = filepath_or_buffer.read()
        except AttributeError:
            # this is a path
            with open(filepath_or_buffer) as fh:
                data = fh.read()
        lines = InifixConf.normalize_data(data)
        if not "".join(lines):
            raise ValueError(f"{filepath_or_buffer} appears to be emtpy.")
        target: Union[InifixParsable, Section] = _dict
        for line_number, line in enumerate(lines, start=1):
            if not line:
                continue
            match = re.match(SECTION_REGEXP, line)
            if match is not None:
                section: Section = _dict[match.group().strip("[]")]
                target = section
                continue

            key, values = InifixConf.tokenize_line(
                line, filename=filepath_or_buffer, line_number=line_number
            )
            if len(values) == 1:
                values = values[0]
            target[key] = values
        super().__init__(_dict)

    @staticmethod
    def normalize_data(data: str) -> List[str]:
        # normalize text body `data` to parsable text lines
        lines = []
        for line in data.splitlines():
            # remove comments and normalize whitespace
            line, _, _comment = line.partition("#")
            lines.append(re.sub(r"\s", " ", line.strip()))
        return lines

    @staticmethod
    def tokenize_line(
        line: str, filename: str, line_number: int
    ) -> Tuple[str, List[Scalar]]:
        key, *raw_values = line.split()
        if not raw_values:
            raise ValueError(f"Failed to parse {filename}:{line_number}:\n{line}")

        values = []
        for val in raw_values:
            # remove period and trailing zeros to cast to int when possible
            val = re.sub(r"\.0+$", "", val)
            casters: List[Callable] = [
                int,
                ENotationIO.decode,
                float,
                bool_caster,
                str_caster,
            ]
            for caster in casters:
                # cast to types from stricter to most permissive
                # `str` will always succeed since it is the input type
                try:
                    values.append(caster(val))
                    break
                except ValueError:
                    continue

        return key, values

    @staticmethod
    def encode(v: Scalar) -> str:
        if isinstance(v, (float, int)):
            return ENotationIO.encode_preferential(v)
        return str(v)

    @staticmethod
    def _write_line(key: str, values: IterableOrSingle[Scalar], buffer: TextIO) -> None:
        val_repr = [InifixConf.encode(v) for v in always_iterable(values)]
        buffer.write(f"{key} {'  '.join([v for v in val_repr])}\n")

    def write_to_buffer(self, buffer: TextIO) -> None:
        for _is_first, is_last, (key, val) in mark_ends(self.items()):
            if not isinstance(val, dict):
                self._write_line(key, val, buffer)
                continue
            buffer.write(f"[{key}]\n")
            for k, v in val.items():
                self._write_line(k, v, buffer)
            if not is_last:
                buffer.write("\n")

    def write_to_file(self, filepath: PathLike) -> None:
        with open(filepath, mode="wt") as fh:
            self.write_to_buffer(fh)

    def write(self, filepath_or_buffer: Union[PathLike, TextIO]) -> None:
        """
        Dump a dict to file in idefix's .ini format.
        """
        try:
            self.write_to_buffer(filepath_or_buffer)
        except AttributeError:
            self.write_to_file(filepath_or_buffer)
